import pygame
from locals import *
import util

class Animation:
    def __init__(self, image, coords):
        self.full_image = util.get_image(image)
        self.full_image.set_colorkey((0, 138, 118)) 
        
        self.images = []
        for c in coords:
            self.images.append(self.full_image.subsurface((c[0],c[1],c[2],c[3])))
            self.images.append(self.images[-1])

class Multiview:
    def __init__(self, image, coords):
        self.full_image = util.get_image(image)
        self.full_image.set_colorkey((0, 138, 118)) 
        
        self.images = {}
        self.rects = {}
        dirs = UL, U, UR, L, R, DL, D, DR

        assert len(dirs) == len(coords)
        
        for d, c in zip(dirs,coords):
            self.images[d] = self.full_image.subsurface((c[0],c[1],c[2],c[3]))
            self.rects[d] = self.images[d].get_rect()

# Explosion
_big_expl = None
_small_expl = None
_fireball = None
_rocket = None
_conversion = None

def init_anims():
    global _big_expl, _small_expl, _fireball, _rocket, _conversion
    _big_expl = Animation("exploBig", 
        ((1,1,39,40),
         (41,1,39,40),
         (121,1,39,40),
         (161,1,39,40),
         (201,1,39,40),
         (241,1,39,40),
         (1,1,39,40),
         (41,41,39,40),
         (121,41,39,40),
         (161,41,39,40),
         (201,41,39,40)))
    _small_expl = Animation("explosn", 
        ((1,1,19,20),                 
        (21,1,19,20),
        (41,1,19,20),
        (61,1,19,20),
        (81,1,19,20),
        (101,1,19,20),
        (121,1,19,20),
        (141,1,19,20)))
    _fireball = Multiview("Bullets",
        ((1,1,9,9),
         (11,1,9,9),
         (21,1,9,9),
         (1,11,9,9),
         (21,11,9,9),
         (1,21,9,9),
         (11,21,9,9),
         (21,21,9,9)))
    _rocket = Multiview("Bullets",
        ((91,1,9,9),
         (101,1,9,9),
         (111,1,9,9),
         (91,11,9,9),
         (101,11,9,9),
         (91,21,9,9),
         (101,21,9,9),
         (111,21,9,9)))
    _conversion = Animation("conversion",
        ((60,0,20,21),
         (100,0,20,21),
         (140,0,20,21),
         (160,0,20,21),
         (180,0,20,21)))
                 
class Explosion(pygame.sprite.Sprite):
    def __init__(self, pos, anim, world):
        pygame.sprite.Sprite.__init__(self)
        self.frame = 0
        self.anim = anim
        self.world = world
        self.image = self.anim.images[self.frame]
        self.rect = self.image.get_rect()
        self.world_pos = pos[0], pos[1]
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        self.rect.left = self.screen_pos[0]
        self.rect.top = self.screen_pos[1]
        self.is_alive = True
        
    def update(self):
        if self.frame < len(self.anim.images)-1:
            self.frame += 1
            self.image = self.anim.images[self.frame]
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]
        else:
            self.kill()

class ScorePoints(pygame.sprite.Sprite):
    def __init__(self, wpos, world, pts, color=None):
        pygame.sprite.Sprite.__init__(self)
        self.world = world
        self.world_pos = wpos
        self.screen_pos = self.world.world_to_screen(wpos)
        self.p = pts
        if color:
            self.image = util.render_text(str(pts), color=color)
        else:
            self.image = util.render_text(str(pts))
        self.rect = self.image.get_rect()
        self.rect.centerx = self.screen_pos[0]
        self.rect.centery = self.screen_pos[1]
        self.frame = 0
        self.max_frames = 16
        
    def update(self):
        if self.frame < self.max_frames:
            self.world_pos[1] -= 2
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.centerx = self.screen_pos[0]
            self.rect.centery = self.screen_pos[1]
            self.frame += 1            
        else:
            self.kill()

class Projectile(pygame.sprite.Sprite):
    # an eight direction based projectile
    def __init__(self, wpos, direc, mview, world):
        pygame.sprite.Sprite.__init__(self)
        self.world = world
        self.world_pos = [wpos[0], wpos[1]]
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        self.v = 7
        self.vx = direc & R and self.v or direc & L and -self.v or 0
        self.vy = direc & D and self.v or direc & U and -self.v or 0  
        self.image = mview.images[direc]
        self.rect = self.image.get_rect()

    def update(self):
        self.world_pos[0] += self.vx 
        self.world_pos[1] += self.vy

        if not self.world.is_onscreen(self.world_pos, self.rect):
            self.kill()
        else:
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]
            
class Projectile2(pygame.sprite.Sprite):
    ### an absolute rotation based projectile
    def __init__(self, wpos, angle, mview, world):
        pygame.sprite.Sprite.__init__(self)
        self.world = world
        self.world_pos = [wpos[0], wpos[1]]
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        v = 7
        self.vx = - v * math.sin(angle)
        self.vy = - v * math.cos(angle)
        self.image = pygame.transform.rotate(mview.images[U], angle * 57.2957795)
        self.rect = self.image.get_rect() 

    def update(self):
        self.world_pos[0] += self.vx
        self.world_pos[1] += self.vy

        if self.world_pos[0] < 0 or self.world_pos[0] > WORLD_WIDTH or \
           self.world_pos[1] < 0 or self.world_pos[1] > WORLD_HEIGHT:
            self.kill()
        else:
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]

def BigExplosion(pos, world):
    global _big_expl
    return Explosion(pos, _big_expl, world)
def SmallExplosion(pos, world):
    global _small_expl
    return Explosion(pos, _small_expl, world)
def Conversion(pos, world):
    global _conversion
    return Explosion(pos, _conversion, world)
def Fireball(wpos, dir, world):
    global _fireball
    return Projectile(wpos, dir, _fireball, world)
def Rocket(wpos, angle, world):
    global _rocket
    return Projectile2(wpos, angle, _rocket, world)
    
