import pygame
from locals import *
import util
import Animation
import random
import Events
import time
from Character import Tank, NPC, Enemy
import sound

class World:
    def __init__(self, screen):
        self.screen = screen
        self.main_char = None
        self.main_char_group = pygame.sprite.Group()
        self.characters = pygame.sprite.Group()
        self.enemies = pygame.sprite.Group()
        self.animations = pygame.sprite.Group()
        self.good_bullets = pygame.sprite.Group()
        self.bad_bullets = pygame.sprite.Group()
        
        self.tile = util.get_image('polar_bg_small', 'jpg')
        self.tile_rect = self.tile.get_rect()

        self.top_left_tile = [0, 0]
        self.min_tl_x = -self.tile_rect.width
        self.min_tl_y = -self.tile_rect.height
        self.x_tiles = (SCREEN_WIDTH / self.tile_rect.width) + 2
        self.y_tiles = (SCREEN_HEIGHT / self.tile_rect.height) + 2
        half_w = SCREEN_WIDTH / 2
        half_h = SCREEN_HEIGHT / 2

        # (left, top, width, height) values for movement bounding box.
        # these values are absolute in relation to the screen, not the world.
        # always use world_to_screen to compare world_pos values to 
        # screen values.
        self.bounding_box = (half_w / 2, half_h / 2, half_w, half_h)
        
        self.world_pos = [0,0]
        self.p_world_pos = [0,0]
        
        self.score = 0
        self.score_img = util.render_score(0)
        self.score_img_pos = (10, SCREEN_HEIGHT-28)
        
        self.lives = LIVES

        self.level_out = 0
        self.game_mode = GAME_NORMAL

    def world_to_screen(self, (wx, wy)):
        return (wx - self.world_pos[0], 
                wy - self.world_pos[1])
                
    def is_onscreen(self, coords, rect):
        sx, sy = self.world_to_screen(coords)
        if sx + rect.width + 2 > 0 and sx < SCREEN_WIDTH and \
           sy + rect.height + 2 > 0 and sy < SCREEN_HEIGHT :
            return True
        return False
        
    def update(self):
        if not hasattr(self, 'main_char'):
            return
        mp = self.world_to_screen(self.main_char.world_pos)
        mr = self.main_char.rect

        self.p_world_pos[0] = self.world_pos[0]
        self.p_world_pos[1] = self.world_pos[1]
    
        # is main_char leaving the bounding box?
        if self.main_char.p_world_pos[0] != self.main_char.world_pos[0]:
            vx = 0
            if mp[0] < self.bounding_box[LEFT]:
                # move left in the world as far as main_char is outside left 
                # edge of bounding box (i.e., keep main_char inside bbox)
                vx = -(self.bounding_box[LEFT] - mp[0])
            elif mp[0] + mr.width > (self.bounding_box[LEFT] + self.bounding_box[WIDTH]):
                vx = mp[0] + mr.width - (self.bounding_box[LEFT] + self.bounding_box[WIDTH])
            self.world_pos[0] += vx
            
            # don't move past the edge of the world
            if self.world_pos[0] < 0:
                vx -= self.world_pos[0] 
                self.world_pos[0] = 0
            elif self.world_pos[0] > (WORLD_WIDTH - SCREEN_WIDTH):
                vx -= self.world_pos[0] - (WORLD_WIDTH - SCREEN_WIDTH)
                self.world_pos[0] = (WORLD_WIDTH - SCREEN_WIDTH)    
            self.top_left_tile[0] -= vx
            if self.top_left_tile[0] < self.min_tl_x:
                self.top_left_tile[0] += self.tile_rect.width
            elif self.top_left_tile[0] > 0:
                self.top_left_tile[0] -= self.tile_rect.width

        if self.main_char.p_world_pos[1] != self.main_char.world_pos[1]:        
            vy = 0
            if mp[1] < self.bounding_box[TOP]:
                vy = -(self.bounding_box[TOP] - mp[1])
            elif mp[1] + mr.height > (self.bounding_box[TOP] + self.bounding_box[HEIGHT]):
                vy = mp[1] + mr.height - (self.bounding_box[TOP] + self.bounding_box[HEIGHT])
            self.world_pos[1] += vy
            if self.world_pos[1] < 0:
                vy -= self.world_pos[1]
                self.world_pos[1] = 0
            elif self.world_pos[1] > (WORLD_HEIGHT - SCREEN_HEIGHT):
                vy -= self.world_pos[1] - (WORLD_HEIGHT - SCREEN_HEIGHT)
                self.world_pos[1] = (WORLD_HEIGHT - SCREEN_HEIGHT)    


            self.top_left_tile[1] -= vy
            if self.top_left_tile[1] < self.min_tl_y:
                self.top_left_tile[1] += self.tile_rect.height
            elif self.top_left_tile[1] > 0:
                self.top_left_tile[1] -= self.tile_rect.height

        self.animations.update()
        self.connections = []
        if self.enemies:
            self.enemies.update()
        elif self.level_out >= LEVEL_OUT:
            self.level_out = 0
            try:
                return Events.trigger("level-up")
            except:
                return Events.trigger("end-game")
        else:
            for c in self.characters:
                c.state = HOLD
            self.level_out += 1
        self.characters.update()
        self.good_bullets.update()
        self.bad_bullets.update()
        self.main_char_group.update()

        gb = pygame.sprite.groupcollide(self.good_bullets, self.enemies, False, False)
        for b in gb:
            if gb[b]:
                gb[b][0].injure(HERO_DMG)
                b.kill()
                self.add_small_explosion(b.world_pos)
                self.add_points(100, b.world_pos)
                sound.play_sound("hit_enemy", .4)
        bb = pygame.sprite.groupcollide(self.bad_bullets, self.characters, False, False)
        for b in bb:
            if bb[b]:
                bb[b][0].kill()
                b.kill()
                self.add_small_explosion(b.world_pos)
                Events.trigger("character-die")
                sound.play_sound("character_explode")
        if pygame.sprite.spritecollide(self.main_char, self.bad_bullets, True):
            self.add_small_explosion(self.main_char.world_pos)
            if not self.main_char.injure(ENEMY_DMG):
                # death event
                self.animations.add(
                    Animation.ScorePoints(self.main_char.world_pos, 
                                          self, "OH NO YOU DIED!", COLOR_RED))
                self.add_explosion((self.main_char.world_pos[0] - 10, self.main_char.world_pos[1]- 10))
                self.add_explosion((self.main_char.world_pos[0] + 10, self.main_char.world_pos[1]+ 10))
                sound.play_sound("hero_explode")
                Events.trigger("hero-die")
            else:
                sound.play_sound("hero_hit")
                self.hp = self.main_char.health
                self.hp_img = util.render_text("hp %03i" % self.hp)
    
    def draw_all(self):
        self.draw_world()
        for c in self.connections:
            pygame.draw.aaline(self.screen, (0,0,0), c[0], c[1])
        self.characters.draw(self.screen)
        self.enemies.draw(self.screen)        
        self.animations.draw(self.screen)
        self.good_bullets.draw(self.screen)
        self.bad_bullets.draw(self.screen)
        self.main_char_group.draw(self.screen)

        # HUD details
        self.screen.blit(self.score_img, self.score_img_pos)
        self.screen.blit(self.hp_img, self.hp_img_pos)
        self.screen.blit(self.lives_img, self.lives_img_pos)
               
    def draw_world(self):
        for yi in xrange(self.y_tiles):
            for xi in xrange(self.x_tiles):
                self.screen.blit(self.tile, 
                    # x value of tile 
                    (self.top_left_tile[0] + (xi * self.tile_rect.width),
                    # y value
                     self.top_left_tile[1] + (yi * self.tile_rect.height)))
   
    def fire_ball(self, char):
        sound.play_sound("hero_shoot")
        self.good_bullets.add(Animation.Fireball(char.world_pos, char.dir, self))
   
    def fire_rocket(self, char, angle):
        sound.play_sound("enemy_shoot", .4)
        self.bad_bullets.add(Animation.Rocket(char.world_pos, angle, self))

    def add_character(self, c):
        self.characters.add(c)
        
    def add_enemy(self, e):
        self.enemies.add(e)
    
    def add_explosion(self, pos):
        self.animations.add(Animation.BigExplosion(pos, self))
    
    def add_small_explosion(self, pos):
        self.animations.add(Animation.SmallExplosion(pos, self))

    def add_conversion(self, pos):
        self.animations.add(Animation.Conversion(pos, self))
        
    def add_points(self, pts, wpos):
        self.animations.add(Animation.ScorePoints(wpos, self, pts))
        self.score += pts
        self.render_score()
        
    def render_score(self):
        self.score_img = util.render_score(self.score)
    
    def respawn(self):
        # respawn the main character
        self.main_char = Tank((self.world_pos[0] + HALF_SCREEN_WIDTH, 
                               self.world_pos[1] + HALF_SCREEN_HEIGHT), self)
        self.main_char_group.add(self.main_char)   
        self.hp = HP
        self.hp_img = util.render_text("hp %03i" % self.hp)
        self.hp_img_pos = (self.score_img.get_rect().width + 30, SCREEN_HEIGHT-28)                          
        if self.lives == 1:
            self.lives_img = util.render_text("lives: %i" % self.lives, color=COLOR_RED)
        else:
            self.lives_img = util.render_text("lives: %i" % self.lives)
        self.lives_img_pos = (self.hp_img_pos[0] + self.hp_img.get_rect().width + 20, 
            SCREEN_HEIGHT-28)
     
def start_level(world, characters, enemies):
    world.main_char_group.empty()
    world.characters.empty()
    world.enemies.empty()
    world.animations.empty()
    world.good_bullets.empty()
    world.bad_bullets.empty()        

    world.char_max = characters
    
    world.respawn()

    sound.play_music("game_song_02")

    for k in KEYS:
        KEYS[k] = False

    for x in range(characters):
        c = NPC('TankRed', (random.randint(0, SCREEN_WIDTH),
                           random.randint(0, SCREEN_HEIGHT)),world)
        world.add_character(c)

    for x in range(enemies):
        e = Enemy('Transport', (random.randint(0, SCREEN_WIDTH),
                           random.randint(0, SCREEN_HEIGHT)),world)
        world.add_enemy(e)    
    
    return world
