"""A generic, keyboard- or joystick-controlled menu module.
Draws a list of strings on the screen, and the user can select one of them.

Key configuration is hard-coded:
* Up/down keys or the main vertical axis on the joystick change the selection
* Return, Z, Space or joystick keys 1 or 2 make the selection.
* Esc quits the menu.

To use the module, initialize the Menu classwith the following parameters:
* screen - standard pygame surface the menu is drawn on
* menu_items - a list of strings in the menu

Optional initialization parameters:
* bgscreen - the menu background covering the entire screen, a pygame surface
* heading_text - a heading for the menu

If the bgscreen isn't specified, the class uses a black screen instead.

After initialization, use the run function. If you want some of the items
pre-selected, you can supply it with the index of the selection in the
menu_items list
    
The run() function returns the index of the selection OR constant MENU_QUIT,
if the user presses Esc while in the menu.

Menu placement on the screen is controlled by constants in the locals.py.
The use of the module also requires play_sound from the sound module,
Util class and render_text from the util module
and the Variables class from the variables module."""

import pygame
import os

from locals import *

import data

from util import Util, render_text
from variables import Variables

from sound import play_sound, play_music

class Menu:

    def __init__(self, screen, menu_items, bgscreen = None, heading_text = None,
       bg_sound=None):
        """Init the menu. Screen is a Pygame Surface, menu_items is a list of
        strings or key, int(value) pairs to display as a menu, bgscreen is the
        menu background Surface and heading_text is a string for the heading."""
        self.bgscreen = bgscreen
        if self.bgscreen == None:
            self.bgscreen = Util.blackscreen
        self.bgscreen = self.bgscreen.copy()
        self.overlay = pygame.Surface((SCREEN_WIDTH, SCREEN_HEIGHT))
        self.overlay.fill((0,0,0))
        self.overlay.set_alpha(160)    
        self.screen = screen
        self.menu_items = []
        self.ret_vals = {}
        for kv in menu_items:
            self.menu_items.append(kv)
            if type(kv) in (tuple, list):
                self.ret_vals[kv[0]] = kv[1]
            else:
                self.ret_vals[kv] = False
        self.bg_sound = bg_sound
        
        if heading_text != None:
            self.heading_text = heading_text
            self.menu_head = render_text(self.heading_text, size=FONT_SIZE_BIG)
            self.rect = self.menu_head.get_rect()
            self.rect.centerx = HALF_SCREEN_WIDTH
            self.menu_offset = -(len(self.menu_items) - 5) * 10
            self.rect.top = GUI_MENU_TOP + 50 + self.menu_offset

    def run(self, menu_choice = 0):
        """Run the menu and return the user's choice."""

        if not self.bg_sound is None:
            play_sound(self.bg_sound)

        done = False
        clock = pygame.time.Clock()

        #Menu loop
        while not done:
            # Pygame event and keyboard input processing
            for event in pygame.event.get():
                if event.type == QUIT or \
                   (event.type == KEYDOWN and event.key == K_ESCAPE):
                    menu_choice = MENU_QUIT
                    self.ret_vals['exit']=True
                    done = True
                elif (event.type == KEYDOWN and event.key == K_DOWN):
                    play_sound("click")
                    if menu_choice + 1 < len(self.menu_items):
                        menu_choice += 1
                    else:
                        menu_choice = 0
                elif (event.type == KEYDOWN and event.key == K_UP):
                    play_sound("click")
                    if menu_choice > 0:
                        menu_choice -= 1
                    else:
                        menu_choice = len(self.menu_items)-1
                elif (event.type == KEYDOWN and event.key == K_RIGHT):
                    if type(self.menu_items[menu_choice]) in (tuple, list):
                        self.ret_vals[self.menu_items[menu_choice][0]] += 1
                elif (event.type == KEYDOWN and event.key == K_LEFT):
                    if type(self.menu_items[menu_choice]) in (tuple, list) and \
                       self.ret_vals[self.menu_items[menu_choice][0]] > 0:
                        self.ret_vals[self.menu_items[menu_choice][0]] -= 1
                elif (event.type == KEYDOWN and \
                   (event.key == K_SPACE or event.key == K_RETURN)):
                    play_sound("click")
                    if not type(self.menu_items[menu_choice]) in (tuple, list):
                        self.ret_vals[self.menu_items[menu_choice]] = True
                        done = True
                    else:
                        if menu_choice + 1 < len(self.menu_items):
                            menu_choice += 1
                        else:
                            menu_choice = 0

            #Renders the menu background, usually the faded out game display
            #Or a black screen
            self.screen.blit(self.bgscreen, (0, 0))
            self.screen.blit(self.overlay, (0, 0))
            
            #Menu rendering
            self.screen.blit(self.menu_head, self.rect)
            for n, mi in enumerate(self.menu_items):
                if type(mi) in (tuple, list):
                    m = "%s  %02i" % (mi[0], self.ret_vals[mi[0]])
                else:
                    m = mi
                
                if n == menu_choice:
                    menu_image = render_text(m, COLOR_GUI_HILIGHT, COLOR_GUI_DARK)
                else:
                    menu_image = render_text(m, COLOR_GUI)
                rect = menu_image.get_rect()
                rect.centerx = HALF_SCREEN_WIDTH
                rect.top = (GUI_MENU_TOP + 120 +
                    (n + 1) * 20 + self.menu_offset)
                self.screen.blit(menu_image, rect)

            pygame.display.flip()
            clock.tick(8)            
        return self.ret_vals
